/*-------------------------------------------------------------------------------
 This file is part of Ranger.
 
 Ranger is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 Ranger is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with Ranger. If not, see <http://www.gnu.org/licenses/>.
 
 Written by:
 
 Marvin N. Wright
 Institut für Medizinische Biometrie und Statistik
 Universität zu Lübeck
 Ratzeburger Allee 160
 23562 Lübeck
 
 http://www.imbs-luebeck.de
 wright@imbs.uni-luebeck.de
#-------------------------------------------------------------------------------*/
 
#include <algorithm>
#include <iostream>
#include <iterator>
#include <vector>
 
#include <Rcpp.h>
 
#include <ctime>
 
#include "utility.h"
#include "TreeRegression.h"
#include "Data.h"
 
 TreeRegression::TreeRegression() :
   counter(0), sums(0) {
 }
 
 TreeRegression::TreeRegression(std::vector<std::vector<size_t>>& child_nodeIDs, std::vector<size_t>& split_varIDs,
                                std::vector<double>& split_values, std::vector<bool>* is_ordered_variable, std::vector<std::vector<double>>& endnodeprobs) :
   Tree(child_nodeIDs, split_varIDs, split_values, is_ordered_variable), counter(0), sums(0), endnodeprobs(endnodeprobs) {
 }
 
 TreeRegression::~TreeRegression() {
   // Empty on purpose
 }
 
 void TreeRegression::initInternal() {
   // Init counters if not in memory efficient mode
   if (!memory_saving_splitting) {
     size_t max_num_unique_values = data->getMaxNumUniqueValues();
     counter = new size_t[max_num_unique_values];
     sums = new double[max_num_unique_values];
   }
 }
 
 double TreeRegression::estimate(size_t nodeID) {
   
   // Mean of responses of samples in node
   double sum_responses_in_node = 0;
   size_t num_samples_in_node = sampleIDs[nodeID].size();
   for (size_t i = 0; i < sampleIDs[nodeID].size(); ++i) {
     sum_responses_in_node += data->get(sampleIDs[nodeID][i], dependent_varID);
   }
   return (sum_responses_in_node / (double) num_samples_in_node);
 }

 std::vector<double> TreeRegression::CalculateProbs(size_t nodeID) {
   
   size_t num_samples_in_node = sampleIDs[nodeID].size();
   size_t nclass = (*borders).size()+1;

   std::vector<size_t> tempsums;
   tempsums.resize(nclass, 0);
   
   for (size_t i = 0; i < num_samples_in_node; ++i) {
     
     // Get the value of the metric response:
     double temp = data->get(sampleIDs[nodeID][i], dependent_varID);
     
     size_t count = 0;
     while (count < nclass-1 && (*borders)[count] < temp) {
       ++count;
     }
     
     // Add this count to the table of class values:
     ++tempsums[count];
     
   }
   
   std::vector<double> probs;
   probs.resize(nclass, 0.0);
   
   for (size_t i = 0; i < nclass; ++i) {
     probs[i] = (double) tempsums[i] / (double) num_samples_in_node;
   }
   
   return probs;
   
 }
 
 void TreeRegression::appendToFileInternal(std::ofstream& file) {
   // Empty on purpose
 }
 
 bool TreeRegression::splitNodeInternal(size_t nodeID, std::vector<size_t>& possible_split_varIDs) {
   
   // Check node size, stop if maximum reached
   if (sampleIDs[nodeID].size() <= min_node_size) {
     split_values[nodeID] = estimate(nodeID);
     endnodeprobs[nodeID] = CalculateProbs(nodeID);
     return true;
   }
   
   // Check if node is pure and set split_value to estimate and stop if pure
   bool pure = true;
   double pure_value = 0;
   for (size_t i = 0; i < sampleIDs[nodeID].size(); ++i) {
     double value = data->get(sampleIDs[nodeID][i], dependent_varID);
     if (i != 0 && value != pure_value) {
       pure = false;
       break;
     }
     pure_value = value;
   }
   if (pure) {
     split_values[nodeID] = pure_value;
     endnodeprobs[nodeID] = CalculateProbs(nodeID);
     return true;
   }
   
   // Find best split, stop if no decrease of impurity
   bool stop;
   if (splitrule == MAXSTAT) {
     stop = findBestSplitMaxstat(nodeID, possible_split_varIDs);
   } else {
     stop = findBestSplit(nodeID, possible_split_varIDs);
   }
   
   if (stop) {
     split_values[nodeID] = estimate(nodeID);
     endnodeprobs[nodeID] = CalculateProbs(nodeID);
     return true;
   }
   
   return false;
 }
 
 void TreeRegression::createEmptyNodeInternal() {
   endnodeprobs.push_back(std::vector<double>());
 }
 
 double TreeRegression::computePredictionAccuracyInternal() {
   
   size_t num_predictions = prediction_terminal_nodeIDs.size();
   double sum_of_squares = 0;
   for (size_t i = 0; i < num_predictions; ++i) {
     size_t terminal_nodeID = prediction_terminal_nodeIDs[i];
     double predicted_value = split_values[terminal_nodeID];
     double real_value = data->get(oob_sampleIDs[i], dependent_varID);
     if (predicted_value != real_value) {
       sum_of_squares += (predicted_value - real_value) * (predicted_value - real_value);
     }
   }
   return (1.0 - sum_of_squares / (double) num_predictions);
 }
 
 double TreeRegression::computePredictionAccuracyInternalCustom() {
   
   if(!userps) {
   
   size_t num_predictions = prediction_terminal_nodeIDs.size();
   size_t num_missclassifications = 0;
   for (size_t i = 0; i < num_predictions; ++i) {
     size_t terminal_nodeID = prediction_terminal_nodeIDs[i];
     double predicted_value = split_values[terminal_nodeID];
     bool stillsmaller = true;
     size_t repsi = 0;
     if (predicted_value > (*borders)[0]) {
       while (stillsmaller && (*borders).size() > repsi + 1) {
         if (predicted_value <= (*borders)[repsi+1]) {
           stillsmaller = false;
         }
         repsi++;
       }
       if((*borders).size() == repsi + 1) {
         repsi++;
       }
       repsi++;
     } else {
       repsi = 1;
     }
     size_t predicted_class = repsi;
     
     double real_value = data->get(oob_sampleIDs[i], dependent_varID);
     
     stillsmaller = true;
     repsi = 0;
     if (real_value > (*borders)[0]) {
       while (stillsmaller && (*borders).size() > repsi + 1) {
         if (real_value <= (*borders)[repsi+1]) {
           stillsmaller = false;
         }
         repsi++;
       }
       if((*borders).size() == repsi + 1) {
         repsi++;
       }
       repsi++;
     } else {
       repsi = 1;
     }
     size_t real_class = repsi;
     
     if (predicted_class != real_class) {
       ++num_missclassifications;
     }
   }
   return (1.0 - (double) num_missclassifications / (double) num_predictions);
   
   } else {
	   
   size_t num_predictions = prediction_terminal_nodeIDs.size();
   double tempsum = 0;
   
   	 size_t nclass = (*borders).size();
	 
        bool stillsmaller = true;
     size_t repsi = 0;
	 	 double indtemp = 0;
	 double cumsumtemp = 0;
   for (size_t i = 0; i < num_predictions; ++i) {
	   
     size_t terminal_nodeID = prediction_terminal_nodeIDs[i];
     std::vector<double> predicted_probs = endnodeprobs[terminal_nodeID];//
     
     double real_value = data->get(oob_sampleIDs[i], dependent_varID);
     
     stillsmaller = true;
     repsi = 0;
     if (real_value > (*borders)[0]) {
       while (stillsmaller && (*borders).size() > repsi + 1) {
         if (real_value <= (*borders)[repsi+1]) {
           stillsmaller = false;
         }
         repsi++;
       }
       if((*borders).size() == repsi + 1) {
         repsi++;
       }
       repsi++;
     } else {
       repsi = 1;
     }
     size_t real_class = repsi;
	 
	 indtemp = 0;
	 cumsumtemp = 0;
	 
	 for (size_t j = 0; j < nclass; ++j) {
		
		if(real_class <= j+1) {
		  indtemp = 1;
		} else {
		  indtemp = 0;
		}
			
		cumsumtemp += predicted_probs[j];
		
		tempsum += (cumsumtemp - indtemp) * (cumsumtemp - indtemp);
		
	 }
	 
	 
   }
   return (- tempsum / (((double) num_predictions) * ((double) nclass - 1)));
   
   }
   
 }
 
 bool TreeRegression::findBestSplit(size_t nodeID, std::vector<size_t>& possible_split_varIDs) {
   
   size_t num_samples_node = sampleIDs[nodeID].size();
   double best_decrease = -1;
   size_t best_varID = 0;
   double best_value = 0;
   
   // Compute sum of responses in node
   double sum_node = 0;
   for (auto& sampleID : sampleIDs[nodeID]) {
     sum_node += data->get(sampleID, dependent_varID);
   }
   
   // For all possible split variables
   for (auto& varID : possible_split_varIDs) {
     
     // Find best split value, if ordered consider all values as split values, else all 2-partitions
     if ((*is_ordered_variable)[varID]) {
       
       // Use memory saving method if option set
       if (memory_saving_splitting) {
         findBestSplitValueSmallQ(nodeID, varID, sum_node, num_samples_node, best_value, best_varID, best_decrease);
       } else {
         // Use faster method for both cases
         double q = (double) num_samples_node / (double) data->getNumUniqueDataValues(varID);
         if (q < Q_THRESHOLD) {
           findBestSplitValueSmallQ(nodeID, varID, sum_node, num_samples_node, best_value, best_varID, best_decrease);
         } else {
           findBestSplitValueLargeQ(nodeID, varID, sum_node, num_samples_node, best_value, best_varID, best_decrease);
         }
       }
     } else {
       findBestSplitValueUnordered(nodeID, varID, sum_node, num_samples_node, best_value, best_varID, best_decrease);
     }
   }
   
   // Stop if no good split found
   if (best_decrease < 0) {
     return true;
   }
   
   // Save best values
   split_varIDs[nodeID] = best_varID;
   split_values[nodeID] = best_value;
   
   // Compute decrease of impurity for this node and add to variable importance if needed
   if (importance_mode == IMP_GINI) {
     addImpurityImportance(nodeID, best_varID, best_decrease);
   }
   return false;
 }
 
 void TreeRegression::findBestSplitValueSmallQ(size_t nodeID, size_t varID, double sum_node, size_t num_samples_node,
                                               double& best_value, size_t& best_varID, double& best_decrease) {
   
   // Create possible split values
   std::vector<double> possible_split_values;
   data->getAllValues(possible_split_values, sampleIDs[nodeID], varID);
   
   // Try next variable if all equal for this
   if (possible_split_values.size() < 2) {
     return;
   }
   
   // Remove largest value because no split possible
   possible_split_values.pop_back();
   
   // Initialize with 0m if not in memory efficient mode, use pre-allocated space
   size_t num_splits = possible_split_values.size();
   double* sums_right;
   size_t* n_right;
   if (memory_saving_splitting) {
     sums_right = new double[num_splits]();
     n_right = new size_t[num_splits]();
   } else {
     sums_right = sums;
     n_right = counter;
     std::fill(sums_right, sums_right + num_splits, 0);
     std::fill(n_right, n_right + num_splits, 0);
   }
   
   // Sum in right child and possbile split
   for (auto& sampleID : sampleIDs[nodeID]) {
     double value = data->get(sampleID, varID);
     double response = data->get(sampleID, dependent_varID);
     
     // Count samples until split_value reached
     for (size_t i = 0; i < num_splits; ++i) {
       if (value > possible_split_values[i]) {
         ++n_right[i];
         sums_right[i] += response;
       } else {
         break;
       }
     }
   }
   
   // Compute decrease of impurity for each possible split
   for (size_t i = 0; i < num_splits; ++i) {
     
     // Stop if one child empty
     size_t n_left = num_samples_node - n_right[i];
     if (n_left == 0 || n_right[i] == 0) {
       continue;
     }
     
     double sum_right = sums_right[i];
     double sum_left = sum_node - sum_right;
     double decrease = sum_left * sum_left / (double) n_left + sum_right * sum_right / (double) n_right[i];
     
     // If better than before, use this
     if (decrease > best_decrease) {
       best_value = possible_split_values[i];
       best_varID = varID;
       best_decrease = decrease;
     }
   }
   
   if (memory_saving_splitting) {
     delete[] sums_right;
     delete[] n_right;
   }
 }
 
 void TreeRegression::findBestSplitValueLargeQ(size_t nodeID, size_t varID, double sum_node, size_t num_samples_node,
                                               double& best_value, size_t& best_varID, double& best_decrease) {
   
   // Set counters to 0
   size_t num_unique = data->getNumUniqueDataValues(varID);
   std::fill(counter, counter + num_unique, 0);
   std::fill(sums, sums + num_unique, 0);
   
   for (auto& sampleID : sampleIDs[nodeID]) {
     size_t index = data->getIndex(sampleID, varID);
     
     sums[index] += data->get(sampleID, dependent_varID);
     ++counter[index];
   }
   
   size_t n_left = 0;
   double sum_left = 0;
   
   // Compute decrease of impurity for each split
   for (size_t i = 0; i < num_unique - 1; ++i) {
     
     // Stop if nothing here
     if (counter[i] == 0) {
       continue;
     }
     
     n_left += counter[i];
     sum_left += sums[i];
     
     // Stop if right child empty
     size_t n_right = num_samples_node - n_left;
     if (n_right == 0) {
       break;
     }
     
     double sum_right = sum_node - sum_left;
     double decrease = sum_left * sum_left / (double) n_left + sum_right * sum_right / (double) n_right;
     
     // If better than before, use this
     if (decrease > best_decrease) {
       best_value = data->getUniqueDataValue(varID, i);
       best_varID = varID;
       best_decrease = decrease;
     }
   }
 }
 
 void TreeRegression::findBestSplitValueUnordered(size_t nodeID, size_t varID, double sum_node, size_t num_samples_node,
                                                  double& best_value, size_t& best_varID, double& best_decrease) {
   
   // Create possible split values
   std::vector<double> factor_levels;
   data->getAllValues(factor_levels, sampleIDs[nodeID], varID);
   
   // Try next variable if all equal for this
   if (factor_levels.size() < 2) {
     return;
   }
   
   // Number of possible splits is 2^num_levels
   size_t num_splits = (1 << factor_levels.size());
   
   // Compute decrease of impurity for each possible split
   // Split where all left (0) or all right (1) are excluded
   // The second half of numbers is just left/right switched the first half -> Exclude second half
   for (size_t local_splitID = 1; local_splitID < num_splits / 2; ++local_splitID) {
     
     // Compute overall splitID by shifting local factorIDs to global positions
     size_t splitID = 0;
     for (size_t j = 0; j < factor_levels.size(); ++j) {
       if ((local_splitID & (1 << j))) {
         double level = factor_levels[j];
         size_t factorID = floor(level) - 1;
         splitID = splitID | (1 << factorID);
       }
     }
     
     // Initialize
     double sum_right = 0;
     size_t n_right = 0;
     
     // Sum in right child
     for (auto& sampleID : sampleIDs[nodeID]) {
       double response = data->get(sampleID, dependent_varID);
       double value = data->get(sampleID, varID);
       size_t factorID = floor(value) - 1;
       
       // If in right child, count
       // In right child, if bitwise splitID at position factorID is 1
       if ((splitID & (1 << factorID))) {
         ++n_right;
         sum_right += response;
       }
     }
     size_t n_left = num_samples_node - n_right;
     
     // Sum of squares
     double sum_left = sum_node - sum_right;
     double decrease = sum_left * sum_left / (double) n_left + sum_right * sum_right / (double) n_right;
     
     // If better than before, use this
     if (decrease > best_decrease) {
       best_value = splitID;
       best_varID = varID;
       best_decrease = decrease;
     }
   }
 }
 
 bool TreeRegression::findBestSplitMaxstat(size_t nodeID, std::vector<size_t>& possible_split_varIDs) {
   
   size_t num_samples_node = sampleIDs[nodeID].size();
   
   // Compute ranks
   std::vector<double> response;
   response.reserve(num_samples_node);
   for (auto& sampleID : sampleIDs[nodeID]) {
     response.push_back(data->get(sampleID, dependent_varID));
   }
   std::vector<double> ranks = rank(response);
   
   // Save split stats
   std::vector<double> pvalues;
   pvalues.reserve(possible_split_varIDs.size());
   std::vector<double> values;
   values.reserve(possible_split_varIDs.size());
   std::vector<double> candidate_varIDs;
   candidate_varIDs.reserve(possible_split_varIDs.size());
   
   // Compute p-values
   for (auto& varID : possible_split_varIDs) {
     
     // Get all observations
     std::vector<double> x;
     x.reserve(num_samples_node);
     for (auto& sampleID : sampleIDs[nodeID]) {
       x.push_back(data->get(sampleID, varID));
     }
     
     // Order by x
     std::vector<size_t> indices = order(x, false);
     //std::vector<size_t> indices = orderInData(data, sampleIDs[nodeID], varID, false);
     
     // Compute maximally selected rank statistics
     double best_maxstat;
     double best_split_value;
     maxstat(ranks, x, indices, best_maxstat, best_split_value, minprop, 1 - minprop);
     //maxstatInData(scores, data, sampleIDs[nodeID], varID, indices, best_maxstat, best_split_value, minprop, 1 - minprop);
     
     if (best_maxstat > -1) {
       // Compute number of samples left of cutpoints
       std::vector<size_t> num_samples_left = numSamplesLeftOfCutpoint(x, indices);
       //std::vector<size_t> num_samples_left = numSamplesLeftOfCutpointInData(data, sampleIDs[nodeID], varID, indices);
       
       // Compute p-values
       double pvalue_lau92 = maxstatPValueLau92(best_maxstat, minprop, 1 - minprop);
       double pvalue_lau94 = maxstatPValueLau94(best_maxstat, minprop, 1 - minprop, num_samples_node, num_samples_left);
       
       // Use minimum of Lau92 and Lau94
       double pvalue = std::min(pvalue_lau92, pvalue_lau94);
       
       // Save split stats
       pvalues.push_back(pvalue);
       values.push_back(best_split_value);
       candidate_varIDs.push_back(varID);
     }
   }
   
   double adjusted_best_pvalue = std::numeric_limits<double>::max();
   size_t best_varID = 0;
   double best_value = 0;
   
   if (pvalues.size() > 0) {
     // Adjust p-values with Benjamini/Hochberg
     std::vector<double> adjusted_pvalues = adjustPvalues(pvalues);
     
     // Use smallest p-value
     double min_pvalue = std::numeric_limits<double>::max();
     for (size_t i = 0; i < pvalues.size(); ++i) {
       if (pvalues[i] < min_pvalue) {
         min_pvalue = pvalues[i];
         best_varID = candidate_varIDs[i];
         best_value = values[i];
         adjusted_best_pvalue = adjusted_pvalues[i];
       }
     }
   }
   
   // Stop if no good split found (this is terminal node).
   if (adjusted_best_pvalue > alpha) {
     return true;
   } else {
     // If not terminal node save best values
     split_varIDs[nodeID] = best_varID;
     split_values[nodeID] = best_value;
     return false;
   }
 }
 
 void TreeRegression::addImpurityImportance(size_t nodeID, size_t varID, double decrease) {
   
   double sum_node = 0;
   for (auto& sampleID : sampleIDs[nodeID]) {
     sum_node += data->get(sampleID, dependent_varID);
   }
   double best_decrease = decrease - sum_node * sum_node / (double) sampleIDs[nodeID].size();
   
   // No variable importance for no split variables
   size_t tempvarID = varID;
   for (auto& skip : *no_split_variables) {
     if (varID >= skip) {
       --tempvarID;
     }
   }
   (*variable_importance)[tempvarID] += best_decrease;
 }
 
 
